@@ -274,48 +274,6 @@ static R_INLINE void hankelize_fft(double *F,
   fftw_free(cV);
 }
 
-static void compute_L_covariation_matrix_first_row(const double *F, R_len_t N, R_len_t L,
-                                                   double *R,
-                                                   const fft_plan *f) {
-  double *oF;
-  fftw_complex *iF, *iFc;
-  R_len_t i, K;
-
-  /* Length's check */
-  if ((L > N) || (L < 1))
-    error("must be 'N' >= 'L' >= 1");
-
-  K = N - L + 1;
-
-  /* Allocate needed memory */
-  oF = (double*) fftw_malloc(N * sizeof(double));
-  iF = (fftw_complex*) fftw_malloc((N/2 + 1) * sizeof(fftw_complex));
-  iFc = (fftw_complex*) fftw_malloc((N/2 + 1) * sizeof(fftw_complex));
-
-  memcpy(oF, F, N*sizeof(double));
-
-  fftw_execute_dft_r2c(f->r2c_plan, oF, iF);
-
-  memcpy(oF, F, K*sizeof(double));
-  memset(oF + K, 0, (N - K)*sizeof(double));
-
-  fftw_execute_dft_r2c(f->r2c_plan, oF, iFc);
-
-  /* Dot-product */
-  for (i = 0; i < N/2 + 1; ++i)
-    iF[i] = iF[i] * conj(iFc[i]);
-
-  fftw_execute_dft_c2r(f->c2r_plan, iF, oF);
-
-  /* Return */
-  for (i = 0; i < L; ++i)
-    R[i] = oF[i] / N;
-
-  /* Cleanup */
-  fftw_free(oF);
-  fftw_free(iF);
-  fftw_free(iFc);
-}
 #else
 static void free_plan(fft_plan *f) {
   (void)f;
@@ -366,7 +324,7 @@ static void hankel_matmul(double* out,
   const hankel_matrix *h = matrix;
   const fft_plan *f = h->fft_plan;
   R_len_t N = h->length, L = h->window;
-  R_len_t K = N - L + 1, i;
+  R_len_t K = h->factor, i;
   SEXP rcirc, rV1, res, rTrue;
 
   /* Allocate needed memory */
@@ -417,7 +375,7 @@ static void hankel_tmatmul(double* out,
   const hankel_matrix *h = matrix;
   const fft_plan *f = h->fft_plan;
   R_len_t N = h->length, L = h->window;
-  R_len_t K = N - L + 1, i;
+  R_len_t K = h->factor, i;
   SEXP rcirc, rV1, res, rTrue;
 
   /* Allocate needed memory */
@@ -527,55 +485,6 @@ static R_INLINE void hankelize_fft(double *F,
   UNPROTECT(6);
 }
 
-static void compute_L_covariation_matrix_first_row(const double *F, R_len_t N, R_len_t L,
-                                                   double *R,
-                                                   const fft_plan *f) {
-  R_len_t K = N - L + 1;
-  R_len_t i;
-  SEXP rF, rFc, rF1, rFc1, rTrue, res;
-
-  if (!valid_plan(f, N))
-    error("invalid FFT plan for given FFT length");
-
-  /* Length check */
-  if ((L > N) || (L < 1))
-    error("must be 'N' > 'L' > 1");
-
-  /* Allocate needed memory */
-  PROTECT(rF = allocVector(CPLXSXP, N));
-  memset(COMPLEX(rF), 0, N * sizeof(Rcomplex));
-  PROTECT(rFc = allocVector(CPLXSXP, N));
-  memset(COMPLEX(rFc), 0, N * sizeof(Rcomplex));
-
-  /* Fill in buffers */
-  for (i = 0; i < N; ++i)
-    COMPLEX(rF)[i].r = F[i];
-
-  for (i = 0; i < K; ++i)
-    COMPLEX(rFc)[i].r = F[i];
-
-  /* Compute the FFTs */
-  PROTECT(rF1 = eval(lang2(install("fft"), rF), R_GlobalEnv));
-  PROTECT(rFc1 = eval(lang2(install("fft"), rFc), R_GlobalEnv));
-
-  /* Dot-product */
-  for (i = 0; i < N; ++i) {
-    Rcomplex x1 = COMPLEX(rF1)[i], x2 = COMPLEX(rFc1)[i];
-    COMPLEX(rF1)[i].r = x1.r * x2.r + x1.i * x2.i;
-    COMPLEX(rF1)[i].i = -x1.r * x2.i + x1.i * x2.r;
-  }
-
-  /* Compute the inverse FFT */
-  PROTECT(rTrue = allocVector(LGLSXP, 1));
-  LOGICAL(rTrue)[0] = 1;
-  PROTECT(res = eval(lang3(install("fft"), rF1, rTrue), R_GlobalEnv));
-
-  /* Form the result */
-  for (i = 0; i < L; ++i)
-    R[i] = COMPLEX(res)[i].r / N;
-
-  UNPROTECT(6);
-}
 #endif
 
 static area_indices *alloc_area1d(SEXP mask) {
@@ -768,88 +677,6 @@ SEXP is_hmat(SEXP ptr) {
   return ans;
 }
 
-SEXP hankel_rows(SEXP ptr) {
-  SEXP tchk;
-  SEXP ans = NILSXP;
-
-  /* Perform a type checking */
-  PROTECT(tchk = is_hmat(ptr));
-
-  if (LOGICAL(tchk)[0]) {
-    ext_matrix *e = R_ExternalPtrAddr(ptr);
-
-    PROTECT(ans = allocVector(INTSXP, 1));
-    INTEGER(ans)[0] = hankel_nrow(e->matrix);
-    UNPROTECT(1);
-  } else
-    error("pointer provided is not a hankel matrix");
-
-  UNPROTECT(1);
-
-  return ans;
-}
-
-SEXP hankel_cols(SEXP ptr) {
-  SEXP tchk;
-  SEXP ans = NILSXP;
-
-  /* Perform a type checking */
-  PROTECT(tchk = is_hmat(ptr));
-
-  if (LOGICAL(tchk)[0]) {
-    ext_matrix *e = R_ExternalPtrAddr(ptr);
-
-    PROTECT(ans = allocVector(INTSXP, 1));
-    INTEGER(ans)[0] = hankel_ncol(e->matrix);
-    UNPROTECT(1);
-  } else
-    error("pointer provided is not a hankel matrix");
-
-  UNPROTECT(1);
-
-  return ans;
-}
-
-SEXP hmatmul(SEXP hmat, SEXP v, SEXP transposed) {
-  SEXP Y = NILSXP, tchk;
-
-  /* Perform a type checking */
-  PROTECT(tchk = is_hmat(hmat));
-
-  if (LOGICAL(tchk)[0]) {
-    R_len_t K, L;
-    ext_matrix *e;
-    hankel_matrix *h;
-
-    /* Grab needed data */
-    e = R_ExternalPtrAddr(hmat);
-    h = e->matrix;
-
-    L = (LOGICAL(transposed)[0] ? hankel_ncol(h) : hankel_nrow(h));
-    K = (LOGICAL(transposed)[0] ? hankel_nrow(h) : hankel_ncol(h));
-
-    /* Check agains absurd values of inputs */
-    if (K != length(v))
-      error("invalid length of input vector 'v'");
-
-    /* Allocate output buffer */
-    PROTECT(Y = allocVector(REALSXP, L));
-
-    /* Calculate the product */
-    if (LOGICAL(transposed)[0])
-      hankel_tmatmul(REAL(Y), REAL(v), h);
-    else
-      hankel_matmul(REAL(Y), REAL(v), h);
-
-    UNPROTECT(1);
-  } else
-    error("pointer provided is not a hankel matrix");
-
-  UNPROTECT(1);
-
-  return Y;
-}
-
 SEXP hankelize_one_fft(SEXP U, SEXP V, SEXP fftplan) {
   /* Perform a type checking */
   if (!LOGICAL(is_fft_plan(fftplan))[0]) {
@@ -918,31 +745,3 @@ SEXP hankelize_multi_fft(SEXP U, SEXP V, SEXP fftplan) {
   UNPROTECT(1);
   return F;
 }
-
-SEXP Lcov_matrix(SEXP F, SEXP L, SEXP fft_plan) {
-  /* Perform a type checking */
-  if (!LOGICAL(is_fft_plan(fft_plan))[0]) {
-    error("pointer provided is not a fft plan");
-    return NILSXP;
-  }
-
-  R_len_t intL = INTEGER(L)[0];
-  R_len_t i, j;
-  R_len_t K = length(F) - intL + 1;
-  SEXP ans;
-
-  PROTECT(ans = allocMatrix(REALSXP, intL, intL));
-  double *rans = REAL(ans);
-  double *pF = REAL(F);
-  compute_L_covariation_matrix_first_row(REAL(F), length(F), intL, rans, R_ExternalPtrAddr(fft_plan));
-
-  for (j = 1; j < intL; ++j)
-    rans[intL*j] = rans[j];
-
-  for (j = 1; j < intL; ++j)
-    for (i = j; i < intL; ++i)
-      rans[i + intL * j] = rans[j + intL * i] = rans[(i - 1) + (j - 1) * intL] - pF[i - 1] * pF[j - 1] + pF[i + K - 1] * pF[j + K - 1];
-
-  UNPROTECT(1);
-  return ans;
-}
