@@ -2,6 +2,7 @@
  *   R package for Singular Spectrum Analysis
  *   Copyright (c) 2009 Konstantin Usevich <usevich.k.d@gmail.com>
  *   Copyright (c) 2009-2010 Anton Korobeynikov <asl@math.spbu.ru>
+ *   Copyright (c) 2014-2015 Alex Shlemov <shlemovalex@gmail.com>
  *
  *   This program is free software; you can redistribute it
  *   and/or modify it under the terms of the GNU General Public
@@ -40,6 +41,7 @@ typedef struct {
   fftw_plan r2c_plan;
   fftw_plan c2r_plan;
 #else
+  Rcomplex *circ_freq;
 #endif
   R_len_t rank;
   R_len_t *window;
@@ -78,6 +80,38 @@ static unsigned hbhankel_ncol(const void *matrix) {
   return h->row_ind != NULL ? h->row_ind->num : prod(h->rank, h->factor);
 }
 
+/* TODO Get rid off this ugly slow complicated function */
+static inline void fill_subarray(double *big,
+                                 double *small,
+                                 R_len_t rank,
+                                 const R_len_t *Nbig,
+                                 const R_len_t *Nsmall,
+                                 int direction) {
+  R_len_t pNsmall = prod(rank, Nsmall);
+  R_len_t *mul, i, j, ii, r;
+
+  mul = Calloc(rank, R_len_t);
+  mul[0] = 1;
+  for (r = 1; r < rank; ++r)
+    mul[r] = mul[r-1] * Nbig[r-1];
+
+  for (i = 0; i < pNsmall; ++i) {
+    ii = i;
+    j = 0;
+    for (r = 0; r < rank; ++r) {
+      j += ii % Nsmall[r] * mul[r];
+      ii /= Nsmall[r];
+    }
+
+    if (direction)
+      big[j] = small[i];
+    else
+      small[i] = big[j];
+  }
+
+  Free(mul);
+}
+
 #if HAVE_FFTW3_H
 static void free_circulant(hbhankel_matrix *h) {
   Free(h->window);
@@ -193,38 +227,6 @@ static void convolveNd(double *x,
   fftw_free(ox);
 }
 
-/* TODO Get rid off this ugly slow complicated function */
-static inline void fill_subarray(double *big,
-                                 double *small,
-                                 R_len_t rank,
-                                 const R_len_t *Nbig,
-                                 const R_len_t *Nsmall,
-                                 int direction) {
-  R_len_t pNsmall = prod(rank, Nsmall);
-  R_len_t *mul, i, j, ii, r;
-
-  mul = Calloc(rank, R_len_t);
-  mul[0] = 1;
-  for (r = 1; r < rank; ++r)
-    mul[r] = mul[r-1] * Nbig[r-1];
-
-  for (i = 0; i < pNsmall; ++i) {
-    ii = i;
-    j = 0;
-    for (r = 0; r < rank; ++r) {
-      j += ii % Nsmall[r] * mul[r];
-      ii /= Nsmall[r];
-    }
-
-    if (direction)
-      big[j] = small[i];
-    else
-      small[i] = big[j];
-  }
-
-  Free(mul);
-}
-
 static void matmul(double* out,
                    const double* v,
                    const void* matrix,
@@ -281,18 +283,6 @@ static void matmul(double* out,
   fftw_free(circ);
 }
 
-static void hbhankel_matmul(double* out,
-                            const double* v,
-                            const void* matrix) {
-  matmul(out, v, matrix, 0);
-}
-
-static void hbhankel_tmatmul(double* out,
-                             const double* v,
-                             const void* matrix) {
-  matmul(out, v, matrix, 1);
-}
-
 static R_INLINE void hbhankelize_fft(double *F,
                                      const double *U, const double *V,
                                      const hbhankel_matrix* h) {
@@ -343,38 +333,401 @@ static R_INLINE void hbhankelize_fft(double *F,
   fftw_free(iU);
   fftw_free(iV);
 }
+
+SEXP convolveN(SEXP x, SEXP y,
+               SEXP input_dim, SEXP output_dim,
+               SEXP Conj) {
+  SEXP x_dim = getAttrib(x, R_DimSymbol);
+  SEXP y_dim = getAttrib(y, R_DimSymbol);
+  R_len_t rank = length(input_dim);
+  R_len_t *N = INTEGER(input_dim);
+  R_len_t pN = prod(rank, N), phN = hprod(rank, N);
+  int conjugate = LOGICAL(Conj)[0];
+
+  fftw_complex *ox, *oy;
+  fftw_plan r2c_plan, c2r_plan;
+  double *circ;
+  R_len_t *revN, r, i;
+
+  /* Allocate needed memory */
+  circ = (double*) fftw_malloc(pN * sizeof(double));
+  ox = (fftw_complex*) fftw_malloc(phN * sizeof(fftw_complex));
+  oy = (fftw_complex*) fftw_malloc(phN * sizeof(fftw_complex));
+
+  /* Estimate the best plans for given input length, note, that input data is
+     stored in column-major mode, that's why we're passing dimensions in
+     *reverse* order */
+  revN = Calloc(rank, R_len_t);
+  for (r = 0; r < rank; ++r) revN[r] = N[rank - 1 - r];
+  r2c_plan = fftw_plan_dft_r2c(rank, revN, circ, ox, FFTW_ESTIMATE);
+  c2r_plan = fftw_plan_dft_c2r(rank, revN, ox, circ, FFTW_ESTIMATE);
+  Free(revN);
+
+  /* Fill input buffer by X values*/
+  memset(circ, 0, pN * sizeof(double));
+  fill_subarray(circ, REAL(x), rank, N, INTEGER(x_dim), 1);
+
+  /* Run the plan on X-input data */
+  fftw_execute_dft_r2c(r2c_plan, circ, ox);
+
+  /* Fill input buffer by Y values*/
+  memset(circ, 0, pN * sizeof(double));
+  fill_subarray(circ, REAL(y), rank, N, INTEGER(y_dim), 1);
+
+  /* Run the plan on Y-input data */
+  fftw_execute_dft_r2c(r2c_plan, circ, oy);
+
+  /* Compute conjugation if needed */
+  if (conjugate)
+    for (i = 0; i < phN; ++i)
+      oy[i] = conj(oy[i]);
+
+  /* Dot-multiply ox and oy, and divide by Nx*...*Nz*/
+  for (i = 0; i < phN; ++i)
+    oy[i] *= ox[i] / pN;
+
+  /* Compute the reverse transform to obtain result */
+  fftw_execute_dft_c2r(c2r_plan, oy, circ);
+
+  SEXP res;
+  PROTECT(res = allocVector(REALSXP, prod(rank, INTEGER(output_dim))));
+  fill_subarray(circ, REAL(res), rank, N, INTEGER(output_dim), 0);
+  /* setAttrib(output_dim, R_NamesSymbol, R_NilValue); */
+  setAttrib(res, R_DimSymbol, output_dim);
+  /* setAttrib(res, R_DimNamesSymbol, R_NilValue); */
+
+  /* Cleanup */
+  fftw_free(ox);
+  fftw_free(oy);
+  fftw_free(circ);
+
+  /* Return */
+  UNPROTECT(1);
+  return res;
+}
 #else
+static void fftn_r2c(const double *z, R_len_t rank, const R_len_t *N,
+                     Rcomplex *res) {
+  SEXP rA, dim, Res;
+  R_len_t n = prod(rank, N);
+
+  rA = PROTECT(allocVector(REALSXP, n));
+  memcpy(REAL(rA), z, sizeof(double) * n);
+
+  dim = PROTECT(allocVector(INTSXP, rank));
+  memcpy(INTEGER(dim), N, sizeof(R_len_t) * rank);
+  setAttrib(rA, R_DimSymbol, dim);
+
+  Res = PROTECT(eval(lang2(install("fft"), rA), R_GlobalEnv));
+
+  /* Return result */
+  memcpy(res, COMPLEX(Res), sizeof(Rcomplex) * n);
+
+  /* Unprotect all */
+  UNPROTECT(3);
+}
+
+static void fftn_c2r(const Rcomplex *z, R_len_t rank, const R_len_t *N,
+                     double *res) {
+  SEXP rTrue, cA, dim, Res;
+  R_len_t n = prod(rank, N), i;
+
+  rTrue = PROTECT(allocVector(LGLSXP, 1));
+  LOGICAL(rTrue)[0] = 1;
+
+  cA = PROTECT(allocVector(CPLXSXP, n));
+  memcpy(COMPLEX(cA), z, sizeof(Rcomplex) * n);
+
+  dim = PROTECT(allocVector(INTSXP, rank));
+  memcpy(INTEGER(dim), N, sizeof(R_len_t) * rank);
+  setAttrib(cA, R_DimSymbol, dim);
+
+  Res = PROTECT(eval(lang3(install("fft"), cA, rTrue), R_GlobalEnv));
+
+  /* Return result */
+  for (i = 0; i < n; ++i)
+    res[i] = COMPLEX(Res)[i].r;
+
+  /* Unprotect all */
+  UNPROTECT(4);
+}
+
 static void free_circulant(hbhankel_matrix *h) {
-  error("FFTW-less version of 2D-SSA is not implemented yet!");
+  Free(h->window);
+  Free(h->factor);
+  Free(h->length);
+
+  Free(h->circ_freq);
 }
 
 static void initialize_circulant(hbhankel_matrix *h,
                                  const double *F,
-                                 R_len_t Nx, R_len_t Ny,
-                                 R_len_t Lx, R_len_t Ly,
+                                 R_len_t rank,
+                                 const R_len_t *N,
+                                 const R_len_t *L,
                                  const int *circular) {
-  error("FFTW-less version of 2D-SSA is not implemented yet!");
+  Rcomplex *ocirc;
+  R_len_t r;
+
+  /* Allocate needed memory */
+  ocirc = Calloc(prod(rank, N), Rcomplex);
+
+  /* Perform FFTn on input data */
+  /* We don't need input buffer here */
+  fftn_r2c(F, rank, N, ocirc);
+
+  /* Return */
+  h->circ_freq = ocirc;
+
+  h->rank = rank;
+
+  h->window = Calloc(rank, R_len_t);
+  memcpy(h->window, L, rank * sizeof(R_len_t));
+
+  h->length = Calloc(rank, R_len_t);
+  memcpy(h->length, N, rank * sizeof(R_len_t));
+
+  h->factor = Calloc(rank, R_len_t);
+  for (r = 0; r < rank; ++r) h->factor[r] = circular[r] ? N[r] : N[r] - L[r] + 1;
 }
 
+static void convolveNd_half(const Rcomplex *ox,
+                            double *y,
+                            R_len_t rank,
+                            const R_len_t *N,
+                            int conjugate) {
+  R_len_t i;
+  Rcomplex *oy;
+  R_len_t pN = prod(rank, N);
+
+  /* Allocate needed memory */
+  oy = Calloc(pN, Rcomplex);
+
+  /* Compute the Nd-FFT of the matrix y */
+  fftn_r2c(y, rank, N, oy);
+
+  /* Compute conjugation if needed */
+  if (conjugate)
+    for (i = 0; i < pN; ++i)
+      oy[i].i = -oy[i].i;
+
+  /* Dot-multiply ox and oy, and divide by Nx*...*Nz*/
+  for (i = 0; i < pN; ++i) {
+    Rcomplex x1 = oy[i], x2 = ox[i];
+    oy[i].r = (x1.r * x2.r - x1.i * x2.i) / pN;
+    oy[i].i = (x1.r * x2.i + x1.i * x2.r) / pN;
+  }
+
+  /* Compute the reverse transform to obtain result */
+  fftn_c2r(oy, rank, N, y);
+
+  /* Cleanup */
+  Free(oy);
+}
+
+static void convolveNd(double *x,
+                       double *y,
+                       R_len_t rank,
+                       const R_len_t *N,
+                       int conjugate) {
+  Rcomplex *ox;
+  R_len_t pN = prod(rank, N);
+
+  /* Allocate needed memory */
+  ox = Calloc(pN, Rcomplex);
+
+  /* Compute the NdFFT of the arrays x and y */
+  fftn_r2c(x, rank, N, ox);
+
+  convolveNd_half(ox, y, rank, N, conjugate);
+
+  /* Cleanup */
+  Free(ox);
+}
+
+static void matmul(double* out,
+                   const double* v,
+                   const void* matrix,
+                   int transposed) {
+  const hbhankel_matrix *h = matrix;
+  R_len_t rank = h->rank;
+  R_len_t *N = h->length;
+  R_len_t *L, *K;
+  R_len_t pN = prod(rank, N);
+  R_len_t i;
+  area_indices *col_ind;
+  area_indices *row_ind;
+  double *circ;
+
+  if (transposed) {
+    K = h->window;
+    L = h->factor;
+    col_ind = h->row_ind;
+    row_ind = h->col_ind;
+  } else {
+    L = h->window;
+    K = h->factor;
+    row_ind = h->row_ind;
+    col_ind = h->col_ind;
+  }
+
+  /* Allocate needed memory */
+  circ = Calloc(pN, double);
+
+  /* Fill the arrays */
+  memset(circ, 0, pN * sizeof(double));
+  if (row_ind == NULL) {
+    fill_subarray(circ, (double*)v, rank, N, K, 1);
+  } else {
+    for (i = 0; i < row_ind->num; ++i) {
+      circ[row_ind->ind[i]] = v[i];
+    }
+  }
+
+  convolveNd_half(h->circ_freq, circ,
+                  rank, h->length,
+                  1);
+
+  /* Cleanup and return */
+  if (col_ind == NULL) {
+    fill_subarray(circ, out, rank, N, L, 0);
+  } else {
+    for (i = 0; i < col_ind->num; ++i) {
+      out[i] = circ[col_ind->ind[i]];
+    }
+  }
+
+  Free(circ);
+}
+
+
+static R_INLINE void hbhankelize_fft(double *F,
+                                     const double *U, const double *V,
+                                     const hbhankel_matrix* h) {
+  R_len_t *N = h->length;
+  R_len_t *L = h->window;
+  R_len_t *K = h->factor;
+  R_len_t rank = h->rank;
+  R_len_t pN = prod(rank, N);
+  R_len_t i;
+
+  double *iU, *iV;
+
+  /* Allocate needed memory */
+  iU = Calloc(pN, double);
+  iV = Calloc(pN, double);
+
+  /* Fill the arrays */
+  memset(iU, 0, pN * sizeof(double));
+  if (h->col_ind == NULL) {
+    fill_subarray(iU, (double*)U, rank, N, L, 1);
+  } else {
+    for (i = 0; i < h->col_ind->num; ++i) {
+      iU[h->col_ind->ind[i]] = U[i];
+    }
+  }
+
+  memset(iV, 0, pN * sizeof(double));
+  if (h->row_ind == NULL) {
+    fill_subarray(iV, (double*)V, rank, N, K, 1);
+  } else {
+    for (i = 0; i < h->row_ind->num; ++i) {
+      iV[h->row_ind->ind[i]] = V[i];
+    }
+  }
+
+  /* Compute convolution */
+  convolveNd(iV, iU, rank, N, 0);
+
+  /* Form the result */
+  for (i = 0; i < pN; ++i) {
+    if (h->weights[i]) {
+      F[i] = iU[i] / h->weights[i];
+    } else {
+      F[i] = NA_REAL;
+    }
+  }
+
+  Free(iU);
+  Free(iV);
+}
+
+SEXP convolveN(SEXP x, SEXP y,
+               SEXP input_dim, SEXP output_dim,
+               SEXP Conj) {
+  SEXP x_dim = getAttrib(x, R_DimSymbol);
+  SEXP y_dim = getAttrib(y, R_DimSymbol);
+  R_len_t rank = length(input_dim);
+  R_len_t *N = INTEGER(input_dim);
+  R_len_t pN = prod(rank, N);
+  int conjugate = LOGICAL(Conj)[0];
+
+  Rcomplex *ox, *oy;
+  double *circ;
+  R_len_t i;
+
+  /* Allocate needed memory */
+  circ = Calloc(pN, double);
+  ox = Calloc(pN, Rcomplex);
+  oy = Calloc(pN, Rcomplex);
+
+  /* Fill input buffer by X values*/
+  memset(circ, 0, pN * sizeof(double));
+  fill_subarray(circ, REAL(x), rank, N, INTEGER(x_dim), 1);
+
+  /* Run the plan on X-input data */
+  fftn_r2c(circ, rank, N, ox);
+
+  /* Fill input buffer by Y values*/
+  memset(circ, 0, pN * sizeof(double));
+  fill_subarray(circ, REAL(y), rank, N, INTEGER(y_dim), 1);
+
+  /* Run the plan on Y-input data */
+  fftn_r2c(circ, rank, N, oy);
+
+  /* Compute conjugation if needed */
+  if (conjugate)
+    for (i = 0; i < pN; ++i)
+      oy[i].i = -oy[i].i;
+
+  /* Dot-multiply ox and oy, and divide by Nx*...*Nz*/
+  for (i = 0; i < pN; ++i) {
+    Rcomplex x1 = oy[i], x2 = ox[i];
+    oy[i].r = (x1.r * x2.r - x1.i * x2.i) / pN;
+    oy[i].i = (x1.r * x2.i + x1.i * x2.r) / pN;
+  }
+
+  /* Compute the reverse transform to obtain result */
+  fftn_c2r(oy, rank, N, circ);
+
+  SEXP res;
+  PROTECT(res = allocVector(REALSXP, prod(rank, INTEGER(output_dim))));
+  fill_subarray(circ, REAL(res), rank, N, INTEGER(output_dim), 0);
+  setAttrib(res, R_DimSymbol, output_dim);
+
+  /* Cleanup */
+  Free(ox);
+  Free(oy);
+  Free(circ);
+
+  /* Return */
+  UNPROTECT(1);
+  return res;
+}
+#endif
+
 static void hbhankel_matmul(double* out,
                             const double* v,
                             const void* matrix) {
-  error("FFTW-less version of 2D-SSA is not implemented yet!");
+  matmul(out, v, matrix, 0);
 }
 
 static void hbhankel_tmatmul(double* out,
                              const double* v,
                              const void* matrix) {
-  error("FFTW-less version of 2D-SSA is not implemented yet!");
+  matmul(out, v, matrix, 1);
 }
 
-static R_INLINE void hbhankelize_fft(double *F,
-                                     const double *U, const double *V,
-                                     const hbhankel_matrix* h) {
-  error("FFTW-less version of 2D-SSA is not implemented yet!");
-}
-#endif
-
 static area_indices *alloc_area2d(SEXP mask, SEXP N) {
   R_len_t r, j, ii;
   if (mask == R_NilValue) {
@@ -508,85 +861,6 @@ SEXP is_hbhmat(SEXP ptr) {
   return ans;
 }
 
-SEXP hbhankel_rows(SEXP ptr) {
-  SEXP tchk;
-  SEXP ans = NILSXP;
-
-  /* Perform a type checking */
-  PROTECT(tchk = is_hbhmat(ptr));
-
-  if (LOGICAL(tchk)[0]) {
-    ext_matrix *e = R_ExternalPtrAddr(ptr);
-
-    PROTECT(ans = allocVector(INTSXP, 1));
-    INTEGER(ans)[0] = hbhankel_nrow(e->matrix);
-    UNPROTECT(1);
-  } else
-    error("pointer provided is not a hankel block-hankel matrix");
-
-  UNPROTECT(1);
-
-  return ans;
-}
-
-SEXP hbhankel_cols(SEXP ptr) {
-  SEXP tchk;
-  SEXP ans = NILSXP;
-
-  /* Perform a type checking */
-  PROTECT(tchk = is_hbhmat(ptr));
-
-  if (LOGICAL(tchk)[0]) {
-    ext_matrix *e = R_ExternalPtrAddr(ptr);
-
-    PROTECT(ans = allocVector(INTSXP, 1));
-    INTEGER(ans)[0] = hbhankel_ncol(e->matrix);
-    UNPROTECT(1);
-  } else
-    error("pointer provided is not a hankel block hankel matrix");
-
-  UNPROTECT(1);
-
-  return ans;
-}
-
-SEXP hbhmatmul(SEXP hmat, SEXP v, SEXP transposed) {
-  SEXP Y = NILSXP, tchk;
-
-  /* Perform a type checking */
-  PROTECT(tchk = is_hbhmat(hmat));
-
-  if (LOGICAL(tchk)[0]) {
-    R_len_t K, L;
-    ext_matrix *e;
-    hbhankel_matrix *h;
-
-    /* Grab needed data */
-    e = R_ExternalPtrAddr(hmat);
-    h = e->matrix;
-
-    L = (LOGICAL(transposed)[0] ? hbhankel_ncol(h) : hbhankel_nrow(h));
-    K = (LOGICAL(transposed)[0] ? hbhankel_nrow(h) : hbhankel_ncol(h));
-
-    /* Check agains absurd values of inputs */
-    if (K != length(v))
-      error("invalid length of input vector 'v'");
-
-    /* Allocate output buffer */
-    PROTECT(Y = allocVector(REALSXP, L));
-
-    /* Calculate the product */
-    matmul(REAL(Y), REAL(v), h, LOGICAL(transposed)[0]);
-
-    UNPROTECT(1);
-  } else
-    error("pointer provided is not a hankel block-hankel matrix");
-
-  UNPROTECT(1);
-
-  return Y;
-}
-
 SEXP hbhankelize_one_fft(SEXP U, SEXP V, SEXP hmat) {
   SEXP F = NILSXP, tchk;
 
@@ -617,75 +891,3 @@ SEXP hbhankelize_one_fft(SEXP U, SEXP V, SEXP hmat) {
 
   return F;
 }
-
-SEXP convolveN(SEXP x, SEXP y,
-               SEXP input_dim, SEXP output_dim,
-               SEXP Conj) {
-  SEXP x_dim = getAttrib(x, R_DimSymbol);
-  SEXP y_dim = getAttrib(y, R_DimSymbol);
-  R_len_t rank = length(input_dim);
-  R_len_t *N = INTEGER(input_dim);
-  R_len_t pN = prod(rank, N), phN = hprod(rank, N);
-  int conjugate = LOGICAL(Conj)[0];
-
-  fftw_complex *ox, *oy;
-  fftw_plan r2c_plan, c2r_plan;
-  double *circ;
-  R_len_t *revN, r, i;
-
-  /* Allocate needed memory */
-  circ = (double*) fftw_malloc(pN * sizeof(double));
-  ox = (fftw_complex*) fftw_malloc(phN * sizeof(fftw_complex));
-  oy = (fftw_complex*) fftw_malloc(phN * sizeof(fftw_complex));
-
-  /* Estimate the best plans for given input length, note, that input data is
-     stored in column-major mode, that's why we're passing dimensions in
-     *reverse* order */
-  revN = Calloc(rank, R_len_t);
-  for (r = 0; r < rank; ++r) revN[r] = N[rank - 1 - r];
-  r2c_plan = fftw_plan_dft_r2c(rank, revN, circ, ox, FFTW_ESTIMATE);
-  c2r_plan = fftw_plan_dft_c2r(rank, revN, ox, circ, FFTW_ESTIMATE);
-  Free(revN);
-
-  /* Fill input buffer by X values*/
-  memset(circ, 0, pN * sizeof(double));
-  fill_subarray(circ, REAL(x), rank, N, INTEGER(x_dim), 1);
-
-  /* Run the plan on X-input data */
-  fftw_execute_dft_r2c(r2c_plan, circ, ox);
-
-  /* Fill input buffer by Y values*/
-  memset(circ, 0, pN * sizeof(double));
-  fill_subarray(circ, REAL(y), rank, N, INTEGER(y_dim), 1);
-
-  /* Run the plan on Y-input data */
-  fftw_execute_dft_r2c(r2c_plan, circ, oy);
-
-  /* Compute conjugation if needed */
-  if (conjugate)
-    for (i = 0; i < phN; ++i)
-      oy[i] = conj(oy[i]);
-
-  /* Dot-multiply ox and oy, and divide by Nx*...*Nz*/
-  for (i = 0; i < phN; ++i)
-    oy[i] *= ox[i] / pN;
-
-  /* Compute the reverse transform to obtain result */
-  fftw_execute_dft_c2r(c2r_plan, oy, circ);
-
-  SEXP res;
-  PROTECT(res = allocVector(REALSXP, prod(rank, INTEGER(output_dim))));
-  fill_subarray(circ, REAL(res), rank, N, INTEGER(output_dim), 0);
-  /* setAttrib(output_dim, R_NamesSymbol, R_NilValue); */
-  setAttrib(res, R_DimSymbol, output_dim);
-  /* setAttrib(res, R_DimNamesSymbol, R_NilValue); */
-
-  /* Cleanup */
-  fftw_free(ox);
-  fftw_free(oy);
-  fftw_free(circ);
-
-  /* Return */
-  UNPROTECT(1);
-  return res;
-}
